
try:
    from logging.handlers import QueueHandler
except ImportError:
    # python < 3.2
    #
    # The class QueueHandler is a direct copy of the implementation of the
    # one from the standard python library.

    import logging

    class QueueHandler(logging.Handler):
        def __init__(self, queue):
            logging.Handler.__init__(self)
            self.queue = queue

        def enqueue(self, record):
            self.queue.put_nowait(record)

        def prepare(self, record):
            self.format(record)
            record.msg = record.message
            record.args = None
            record.exc_info = None
            return record

        def emit(self, record):
            try:
                self.enqueue(self.prepare(record))
            except Exception:
                self.handleError(record)


__all__ = ['QueueHandler']



import threading

import six


class CounterMetric(object):
    def __init__(self):
        self._value = 0
        self._count = 0

    def inc(self, value=1):
        self._value += value
        self._count += 1

    def value(self):
        return self._value

    def reset(self):
        self._value = 0
        self._count = 0

    def avg(self):
        if self._count == 0:
            return 0
        return float(self._value) / float(self._count)


class MetricCollection(object):
    def __init__(self):
        self._lock = threading.RLock()
        self._metrics = {}
        self._has_metrics = False

    def inc(self, metric_key, value=1):
        assert isinstance(metric_key, (six.binary_type, six.text_type))
        assert isinstance(value, int), value
        with self._lock:
            self._has_metrics = True
            metric = self._metrics.get(metric_key)
            if not metric:
                metric = CounterMetric()
                self._metrics[metric_key] = metric
            metric.inc(value)

    def has_metrics(self):
        return self._has_metrics

    def get_all_and_reset(self):
        retval = {}
        with self._lock:
            self._has_metrics = False
            for metric_key, metric in six.iteritems(self._metrics):
                retval[metric_key] = metric.value()
                retval[metric_key + '.avg'] = metric.avg()

                metric.reset()

        return retval



import logging
import os
import warnings


class BaseMultiprocessHandler(logging.Handler):
    def __init__(self, logger):
        logging.Handler.__init__(self)

        self._parent_pid = os.getpid()
        self._sink_logger = logging.getLogger(logger)

    def emit(self, record):
        """If called by a subprocess, then the log record is send to the queue.

        If called by the main process, then the log record is forwarded to the
        the logger ``starlog.logsink`` or whatever was passed to the ``logger``
        parameter at instantiation of the MultiprocessHandler.
        """
        if self._is_main_process():
            self.forward_to_sink(record)
        else:
            self.forward_to_main(record)

    def _is_main_process(self):
        return self._parent_pid == os.getpid()

    def forward_to_sink(self, record):
        if not self._sink_logger.handlers:
            msg = 'The logger %s does not have any handlers configured' % (
                self._sink_logger.name)
            warnings.warn(msg, UserWarning)

        self._sink_logger.handle(record)

    def forward_to_main(self, record):
        raise NotImplementedError()



import json
from logging import makeLogRecord, NullHandler
try:
    import cPickle as pickle
except ImportError:
    # python 3
    import pickle


null_handler = NullHandler()


def record_to_dict(record, handler=null_handler):
    # this function is the core of python's logging.handler.SocketHandler
    # method ``makePickle``

    ei = record.exc_info
    if ei:
        # just to get traceback text into record.exc_text ...
        handler.format(record)
    # See issue #14436: If msg or args are objects, they may not be
    # available on the receiving end. So we convert the msg % args
    # to a string, save it as msg and zap the args.
    d = dict(record.__dict__)
    d['msg'] = record.getMessage()
    d['args'] = None
    d['exc_info'] = None
    # Issue #25685: delete 'message' if present: redundant with 'msg'
    d.pop('message', None)
    return d


def pickle_log_record(record, handler=null_handler):
    """Pickles the record in binary format with a length prefix.
    """
    d = record_to_dict(record, handler=handler)
    s = pickle.dumps(d, 1)
    return s


def unpickle_log_record(data):
    """Creates a LogRecord from a pickled LogRecord.

    :param bytes data: pickled version of a LogRecord created by
        ``pickle_log_record``
    :return: a LogRecord
    """
    unpickled = pickle.loads(data)
    return makeLogRecord(unpickled)


def json_encode_log_record(record, handler=null_handler):
    d = record_to_dict(record, handler=null_handler)
    return json.dumps(d)


def json_decode_log_record(data):
    d = json.loads(data)
    return makeLogRecord(d)



import errno
import os
import re
import threading
import traceback
from logging import makeLogRecord

import zmq

from .base_handler import BaseMultiprocessHandler
from .debug import get_debug_logger
from .serializer import record_to_dict
from .utils import retry, RetryAbortedByCheck


_log = get_debug_logger('starlog.debug.zmq_handler')


class BindFailedError(Exception):
    pass


def _requires_random_bind(address):
    if not address.startswith('tcp://'):
        return False

    if re.search(r':\d+$', address):
        return False
    else:
        return True


class RobustZmqSocket(object):
    UNRECOVERABLE_ERRORS = [
        # "Permission denied"
        errno.EACCES,
    ]

    DEFAULT_SOCKET_OPTIONS = {
        # set recv timeout
        'RCVTIMEO': 1000,
        'LINGER': 0,
    }

    def __init__(self, socket_type, address, backoff_factor=2.0, tries=8,
                 check=None, socket_options=DEFAULT_SOCKET_OPTIONS):
        # default: tries up to 4 minutes 15 seconds to bind / connect to
        # a socket
        self._socket_type = socket_type
        self._address = address

        self._context = None
        self._socket = None
        self._socket_options = socket_options

        self._retry_bind = retry(
            zmq.ZMQError,
            tries=tries,
            backoff_factor=backoff_factor,
            check=check,
            retry_log=self._log_bind_attempt)

        self._retry_connect = retry(
            zmq.ZMQError,
            tries=tries,
            backoff_factor=backoff_factor,
            check=check,
            retry_log=self._log_bind_attempt)

    def _obtain_context(self):
        if self._context is None:
            self._context = zmq.Context()

        return self._context

    def _log_bind_attempt(self, _trial, last_trial=False):
        if last_trial:
            _log.error('bind to %s failed. Aborting.', self._address)
        else:
            _log.warning('bind to %s failed. Retrying', self._address)

    def _log_connect_attempt(self, _trial, last_trial=False):
        if last_trial:
            _log.error('connect to %s failed. Aborting.', self._address)
        else:
            _log.warning('connect to %s failed. Retrying.', self._address)

    def bind(self):
        bind_with_retries = self._retry_bind(self._bind)
        bind_with_retries()

    def _bind(self):
        try:
            self.close_socket()
            context = self._obtain_context()
            self._socket = context.socket(self._socket_type)

            if _requires_random_bind(self._address):
                port = self._socket.bind_to_random_port(self._address)
                self._address = '%s:%d' % (self._address, port)
            else:
                self._socket.bind(self._address)

            self._set_socket_options()
        except zmq.ZMQError as error:
            if error.errno in self.UNRECOVERABLE_ERRORS:
                raise BindFailedError(str(error))

            raise

    def connect(self):
        connect_with_retries = self._retry_connect(self._connect)
        connect_with_retries()

    def _connect(self):
        self.close_socket()
        context = self._obtain_context()
        self._socket = context.socket(self._socket_type)
        self._socket.connect(self._address)

    def _set_socket_options(self):
        if not self._socket_options:
            return

        for option, value in self._socket_options.items():
            setattr(self._socket, option, value)

    def recv_json(self):
        try:
            return self._socket.recv_json()
        except zmq.ZMQError as error:
            if error.errno == errno.EAGAIN:
                # recv timeout
                return

            self.bind()

    def send_json(self, data):
        try:
            return self._socket.send_json(data)
        except zmq.ZMQError:
            self.connect()
            return self._socket.send_json(data)

    def close(self):
        """Close the zmq socket and destroy the zmq context.
        """
        self.close_socket()
        self.destroy_context()

    def destroy_context(self):
        """Destroy the zmq context.
        """
        if self._context is not None:
            if not self._context.closed:
                self._context.destroy()
            self._context = None

    def close_socket(self):
        """Close the zmq socket.
        """
        if self._socket is not None:
            if not self._socket.closed:
                self._socket.close()
            self._socket = None

    @property
    def address(self):
        assert not _requires_random_bind(self._address)
        return self._address


class ZmqHandler(BaseMultiprocessHandler):
    """The ZmqHandler creates a zmq connection between the main
    process and its children processes. The main process establishes a
    :py:const:`zmq.PULL` socket. Child processes sets up a :py:const:`zmq.PUSH`
    socket.

    All subprocesses sends messages to the zmq connection in the `emit`
    method. A background thread of the main process retrieves these messages
    and delegates them to the ``logger`` (default: ``starlog.sink``) in the
    main process.

    The ZmqHandler is expected to be set up by the main process.

    :param str address: the address of the connection which is passed to
        :py:meth:`zmq.Socket.connect` / :py:meth:`zmq.Socket.bind`.
        Examples: ``tcp://127.0.0.1:12345``, ``tcp://127.0.0.1``,
        ``ipc:///tmp/log.sock``. Omitting the port in a ``tcp://`` address
        will bind the socket to a random port.
    :param str logger: name of the logger where log records are send to in the
        main process.
    """

    def __init__(self, address='tcp://127.0.0.1:5557',
                 logger='starlog.logsink'):
        BaseMultiprocessHandler.__init__(self, logger)

        self._sender_socket_type = zmq.PUSH
        self._receiver_socket_type = zmq.PULL

        self._address = address
        self._async_listener = self._start_listener_thread(address)

        # for children processes
        self._pid = None

    def _start_listener_thread(self, address):
        _log.info('ZmqHandler._start_listener_thread')

        event = threading.Event()
        listener = ZmqListenerThread(self._receiver_socket_type, address,
                                     event, self)
        listener.start()

        # wait until the socket is bound in the listener
        for i in range(60):
            event.wait(1)
            if event.is_set():
                break

            if not listener.is_alive():
                break

        if not event.is_set():
            raise Exception('Listener thread could not be created')

        self._address = listener.get_address()

        return listener

    def forward_to_main(self, record):
        socket = self._get_socket()
        record_dict = record_to_dict(record, self)
        socket.send_json(record_dict)

    def _get_socket(self):
        pid = os.getpid()

        if self._pid != pid:
            # pid is None: new child forked from main
            # pid != os.getpid(): new child forked from other child
            self._client_bind_to_socket()
            self._pid = pid

        return self._socket

    def _client_bind_to_socket(self):
        socket = RobustZmqSocket(self._sender_socket_type, self._address)
        socket.connect()

        self._socket = socket

    def _close_socket(self):
        try:
            socket = self._socket
        except AttributeError:
            return
        else:
            if socket is not None:
                socket.close()
                self._socket = None

    def close(self):
        _log.info('ZmqHandler.close for %s', self)
        BaseMultiprocessHandler.close(self)

        try:
            if not self._is_main_process():
                return

            listener_thread = self._async_listener
            if listener_thread is None:
                return

            listener_thread.shutdown()
            listener_thread.join()
        finally:
            self._close_socket()


class ZmqListenerThread(threading.Thread):
    def __init__(self, socket_type, address, event, handler, *args, **kwargs):
        super(ZmqListenerThread, self).__init__(*args, **kwargs)
        self.setDaemon(True)

        self._event = event
        self._handler = handler
        self._running = True

        self._zmq_socket = RobustZmqSocket(
            socket_type, address, check=self._not_running)

    def _not_running(self):
        return not self._running

    def run(self):
        try:
            self._zmq_socket.bind()
            self._event.set()
            self._event = None

            while self._running:
                record_dict = self._zmq_socket.recv_json()
                if record_dict is None:
                    # receive timeout
                    continue

                self._process_record(record_dict)
        except RetryAbortedByCheck:
            # zmq bind failed and thread stopped running
            pass
        except Exception:
            _log.warning('exception in %s.run: %s',
                         self.__class__.__name__, traceback.format_exc())

        _log.info('%s stopped', self.__class__.__name__)
        self.close()

    def _process_record(self, record_dict):
        record = makeLogRecord(record_dict)
        self._handler.forward_to_sink(record)

    def get_address(self):
        return self._zmq_socket.address

    def close(self):
        """Close the zmq socket connection.
        """
        self._zmq_socket.close()

    def shutdown(self):
        """Gracful shutdown.
        """
        self._running = False



"""For debugging the starlog library.

To enable debugging you must set the environment variable ``STARLOG_DEBUG=1``
or via calling ``starlog.log_to_stderr()`` at the beginning of your code.
"""
import logging
import os
import sys


_logging_configured = False

STARLOG_DEBUG = os.environ.get('STARLOG_DEBUG', '0') == '1'


def get_debug_logger(logger_name):
    '''Returns the logger to debug starlog.
    '''
    assert logger_name.startswith('starlog.debug')

    _pre_configure_debug_logging()

    return logging.getLogger(logger_name)


def log_to_stderr(level=logging.DEBUG):
    global STARLOG_DEBUG, _logging_configured

    STARLOG_DEBUG = 1

    _logging_configured = False
    _pre_configure_debug_logging(level=level)


def _pre_configure_debug_logging(level=logging.DEBUG):
    """Do a pre configuration here at a very early stage even before
    logging.basicConfig or logging.fileConfig is called.
    """
    global _logging_configured

    if _logging_configured:
        return

    logger = logging.getLogger('starlog.debug')

    if not STARLOG_DEBUG:
        logger.disabled = True
        logger.setLevel(logging.CRITICAL)
    else:
        logger.info('starlog debug log activated')
        if not logger.handlers and logger.propagate:
            formatter = logging.Formatter('[%(name)s:%(process)d] %(message)s')
            handler = logging.StreamHandler(sys.stderr)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            logger.propagate = 0
            logger.setLevel(level)

    _logging_configured = True



from .debug import log_to_stderr
from .log_entry import inc
from .multiprocess_handler import MultiprocessHandler
from .status_handler import StatusHandler


def replace_callable_with_import_error(message):
    def wrap(*args, **kwargs):
        raise ImportError(message)
    return wrap


try:
    from .zmq_handler import ZmqHandler
except ImportError as error:
    message = '%s. Please install "starlog[zmq]"' % (error, )
    ZmqHandler = replace_callable_with_import_error(message)


__all__ = [
    'MultiprocessHandler',
    'StatusHandler',
    'ZmqHandler',
    'inc',
    'log_to_stderr',
    ]



import multiprocessing
import threading
import traceback

from .base_handler import BaseMultiprocessHandler
from .compat import QueueHandler
from .debug import get_debug_logger
from .utils import sleep


_log = get_debug_logger('starlog.debug.queue_handler')


class MultiprocessHandler(BaseMultiprocessHandler):
    """The MultiprocessHandler uses :py:class:`multiprocessing.Queue` to send
    log records between processes.

    All subprocesses sends messages to the queue in the `emit`
    method. A background thread of the main process retrieves these messages
    and delegates them to the ``logger`` (default: ``starlog.sink``) in the
    main process.

    The MultiprocessHandler is expected to be set up by the main process.

    :param queue: A multiprocessing capable queue. If not set, then a new
        multiprocessing queue is created.
    :param bool manager_queue: If `queue` is `None` and this argument is set
        to `True`, then a new queue is created by calling
        :py:meth:`multiprocessing.managers.SyncManager.Queue`.
    :param str logger: name of the logger where log records are send to in the
        main process.
    """

    def __init__(self, queue=None, manager_queue=True,
                 logger='starlog.logsink'):
        BaseMultiprocessHandler.__init__(self, logger)

        if queue is None:
            if manager_queue:
                manager = multiprocessing.Manager()
                queue = manager.Queue()
            else:
                queue = multiprocessing.Queue()
        else:
            self._check_queue(queue)

        self._queue = queue
        self._qhandler = QueueHandler(queue)

        self._async_listener = self._start_listener_thread(queue)

    def _check_queue(self, queue):
        """Check if ``queue`` has methods .get and .put
        """
        try:
            queue.get
            queue.put
        except AttributeError:
            raise TypeError('queue %r is of unsupported type: %s' % (
                queue, queue.__class__))

    def _start_listener_thread(self, queue):
        _log.info('MultiprocessHandler._start_listener_thread')
        listener = QueueListenerThread(queue, self)
        listener.start()
        return listener

    def forward_to_main(self, record):
        self._qhandler.handle(record)

    def close(self):
        _log.info('MultiprocessHandler.close for %s', self)
        BaseMultiprocessHandler.close(self)

        if not self._is_main_process():
            return

        self._shutdown_listener()

    def _shutdown_listener(self):
        listener_thread = self._async_listener
        if listener_thread is None:
            return

        queue = self._queue
        if queue is None:
            return

        # send signal to shutdown the QueueListenerThread
        try:
            queue.put(None)
        except (IOError, EOFError) as error:
            # probably the queue is already closed
            _log.warning('error while closing QueueListenerThread: %s', error)
            raise

        # wait until the thread consumed the shutdown item
        if not sleep(10, check=listener_thread.is_alive):
            _log.warning('QueueListenerThread did not shut down')

        _close_queue(queue)
        _join_queue(queue)
        self._queue = None
        self._async_listener = None


def _close_queue(queue):
    try:
        queue.close
    except AttributeError:
        pass
    else:
        _log.info('calling queue.close()')
        queue.close()
        _log.info('queue closed')


def _join_queue(queue):
    try:
        queue.join
    except AttributeError:
        pass
    else:
        _log.info('calling queue.join()')
        queue.join()
        _log.info('queue joined')


def _task_done(queue):
    try:
        queue.task_done
    except AttributeError:
        pass
    else:
        _log.info('QueueListenerThread: calling queue.task_done()')
        try:
            queue.task_done()
        except IOError:
            pass


class QueueListenerThread(threading.Thread):
    """Listens for incoming queue messages and forwards them back to the
    corresponding logger, i.e. in the main process.
    """
    def __init__(self, queue, handler, *args, **kwargs):
        super(QueueListenerThread, self).__init__(*args, **kwargs)
        self.setDaemon(True)

        self._queue = queue
        self._handler = handler

    def run(self):
        """Start the main loop. Wait for incoming log records in the queue.
        Each log record is forwarded back to python's standard logger of the
        main process.
        """
        try:
            while True:
                try:
                    # A closed queue results in an IOError in the get() request
                    # only if no timeout is set. A None value in the queue is
                    # a signal to stop the thread.
                    record = self._queue.get()
                    if record is None:
                        _log.info('shutting down QueueListenerThread')
                        break
                    self._process_record(record)
                except (IOError, EOFError):
                    # multiprocessing.managers.Queue returns an EOFError
                    _log.info('exception in QueueListenerThread.run: %s',
                              traceback.format_exc())
                    break
        except Exception:
            _log.warning('exception in QueueListenerThread.run: %s',
                         traceback.format_exc())

        _log.info('QueueListenerThread stopped')

        _task_done(self._queue)

    def _process_record(self, record):
        self._handler.forward_to_sink(record)



EXTRA_KEY = 'STARLOG_EXTRA'


class Extra(object):
    def __init__(self):
        self._dic = {}

    def update(self, dic):
        self._dic.update(dic)
        return self._dic

    def inc(self, metric_key, value=1):
        if EXTRA_KEY not in self._dic:
            self._dic[EXTRA_KEY] = {}

        self._dic[EXTRA_KEY][metric_key] = value
        return self

    def items(self):
        return self._dic.items()

    def __iter__(self):
        return iter(self.items())


def inc(metric_key, value=1, dic=None):
    """For use in the `extra` parameter of logger.log() like

    logger.info('incoming request', extra=inc('requests'))

    logger.info('job done',
                extra=inc('jobs-done').inc('commits'))

    logger.info('some more meta data',
                extra=inc('changes').update({'caller': 'joe'}))
    """
    if dic is None:
        dic = Extra()

    dic.inc(metric_key, value)

    return dic


def get_log_record_metric(log_record):
    """Extract metric from log record.
    """
    return log_record.__dict__.get(EXTRA_KEY)



"""A log handler that examines every log and does some statistics on it.
It generates a log message in regular intervals.

Metrics that are collected:

- the number of messages per log level, available in formatter as:

  - ``%(DEBUG)d``
  - ``%(INFO)d``
  - ``%(WARN)d``
  - ``%(ERROR)d``
  - ``%(FATAL)d``

- the size of messages per log level, available in formatter as

  - ``%(DEBUG-SIZE)d``
  - ``%(INFO-SIZE)d``
  - ``%(WARN-SIZE)d``
  - ``%(ERROR-SIZE)d``
  - ``%(FATAL-SIZE)d``

- custom values examples::

    from starlog import inc
    logger.info('one more log line', extra=inc('requests').inc('2xx'))
    logger.info('bad request', extra=inc('4xx').update(error='Invalid input'))

  - ``%(requests)d``
  - ``%(2xx)d``
  - ``%(4xx)d``


Use case: write out status logs every some seconds to standard out and sent
full logs to syslog or a file.
"""
import logging
import logging.handlers
import re
import threading
import time
import traceback
import warnings
from collections import defaultdict

from .debug import get_debug_logger
from .log_entry import get_log_record_metric
from .metrics import MetricCollection


metric_collection = MetricCollection()

_log = get_debug_logger('starlog.debug.status_handler')


def _parse_interval(value):
    if not isinstance(value, int):
        match = re.match(r'(\d+)([smh]?)', value, flags=re.I)
        if not match:
            raise ValueError('interval is in wrong format: %r' % (value,))

        value = int(match.group(1))
        unit = match.group(2)

        if unit == 'm':
            value *= 60
        elif unit == 'h':
            value *= 3600

    if value < 0:
        raise ValueError('interval must be greater zero. Got: %r' % (value,))

    if value == 0:
        warnings.warn(
            'interval is zero. This could lead to high cpu consumption')

    return value


class StatusHandler(logging.Handler):
    """StatusHandler is a log handler that aggregates log messages and
    generates a status log message in regular intervals.

    :param str interval: the log status interval. Values must be of the format
        **Xs** (seconds), **Xm** (minutes) or **Xh** (hours), where **X** is
        a positive integer value. If you pass an integer instead of a string,
        then this value is taken as the number of seconds.
        Examples: **15s**, **5m**, **1h**
    :type interval: str or int

    Example::

        import logging
        import starlog
        import sys
        import time


        formatter = logging.Formatter('Seen info messages: %(INFO)d')
        handler = starlog.StatusHandler('1s')
        logging.root.addHandler(handler)
        logging.root.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler(sys.stdout)
        stdout.setFormatter(formatter)
        logging.getLogger('starlog.status').addHandler(stdout)

        logging.root.info('you will not see this message - just the status')

        time.sleep(5)

        # output is:
        >>> Seen info messages: 1
        >>> Seen info messages: 0
        >>> Seen info messages: 0
        >>> Seen info messages: 0
        >>> Seen info messages: 0
    """
    def __init__(self, interval='5s', logger='starlog.status'):
        logging.Handler.__init__(self)
        self._metrics = metric_collection

        self._logger_name = logger
        self._async_reporter = self._start_reporter_thread(
            interval=interval,
            logger=logging.getLogger(logger))

    def emit(self, record):
        """Aggregates and collects metrics about the log record.
        The actual log record does not generate any direct output with this log
        handler. For this you must configure the logger ``starlog.status``
        """
        if record.name == self._logger_name:
            # if status logger was not configured with 'propagate=0'
            return
        self._count_messages(record)
        self._count_custom_metrics(record)

    def _count_messages(self, record):
        levelname = record.levelname or 'UNKNOWN'
        size = len(record.msg)
        self._metrics.inc(levelname)
        self._metrics.inc(levelname + '-SIZE', size)

    def _count_custom_metrics(self, record):
        metric = get_log_record_metric(record)
        if not metric:
            return

        for metric_key, value in metric.items():
            self._metrics.inc(metric_key, value)

    def _start_reporter_thread(self, **kwargs):
        _log.info('StatusHandler._start_reporter_thread')
        reporter_thread = ReporterThread(self._metrics, **kwargs)
        reporter_thread.start()
        return reporter_thread

    def close(self):
        _log.info('StatusHandler.close')
        logging.Handler.close(self)

        reporter_thread = self._async_reporter

        if reporter_thread is not None:
            reporter_thread.shutdown()
            reporter_thread.join()


class ReportLogRecord(object):
    def __init__(self, record):
        object.__setattr__(self, '_record', record)

    @property
    def __dict__(self):
        d = self._record.__dict__
        return defaultdict(lambda: 0, d)

    def __getattr__(self, key):
        return getattr(self._record, key)

    def __setattr__(self, key, value):
        setattr(self._record, key, value)


class ReporterThread(threading.Thread):
    def __init__(self, metrics, interval, logger, *args, **kwargs):
        super(ReporterThread, self).__init__(*args, **kwargs)
        self.setDaemon(True)

        self._metrics = metrics
        self._interval = _parse_interval(interval) * 2
        self._logger = logger
        self._running = True

    def shutdown(self):
        """Graceful shutdown.
        """
        _log.info('shutting down ReporterThread')
        self._running = False

    def run(self):
        try:
            while self._running:
                self._sleep()
                if self._running:
                    self._log_status()

            if self._metrics.has_metrics():
                self._log_status()
        except Exception:
            _log.warn('in ReporterThread.run: %s', traceback.format_exc())

    def _log_status(self):
        metrics = self._metrics.get_all_and_reset()

        record = logging.LogRecord(
            self._logger.name, logging.INFO, "", 0, "", (), None, None)
        record.__dict__.update(metrics)
        report_record = ReportLogRecord(record)

        self._logger.handle(report_record)

    def _sleep(self):
        for i in range(self._interval):
            if not self._running:
                break
            time.sleep(0.5)



import os
from setuptools import setup


here = os.path.dirname(os.path.abspath(__file__))
long_description = open(os.path.join(here, 'README.md')).read()

setup(name='starlog',
      version='1.0.0',
      description='Advanced logging features.',
      author='Michael Peick',
      author_email='michael.peick+starlog@gmail.com',
      url='https://gitlab.com/peick/starlog',
      license='gpl-3.0',
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords=['logging', 'log handler', 'status logging',
                'multiprocessing', 'zmq'],
      packages=['starlog'],
      install_requires=[
          'six',
      ],
      extras_require={
          'zmq': [
              'pyzmq',
          ],
      },
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: System :: Logging',
      ]
      )



import time


class RetryAbortedByCheck(Exception):
    pass


def sleep(duration, check=None, step=1.0):
    end = time.time() + duration
    rest = duration
    while rest > 0:
        if check is not None and check():
            return True

        if rest > step:
            time.sleep(step)
            rest = end - time.time()
        elif rest <= 0:
            return
        else:
            time.sleep(rest)
            return


def retry(exceptions=Exception, tries=3, backoff_factor=1, check=None,
          retry_log=None):
    """
    :param float backoff_factor: time sleep between retires is calculated with
        ``duration = backoff_factor * 2 ** (retries - 1)``
    :param callable check: A function / callable that is called on before every
        retry attempt. If value returned by ``check()`` evaluates to true,
        then ``retry`` raises a ``RetryAbortedByCheck`` exception.
    :param callable retry_log: A function / callable that is called before
        every retry attempt and on the last trial error. The function must
        accept the arguments ``retry_log(trial, last_trial=False)``
    """
    assert tries >= 0
    assert backoff_factor >= 0

    def wrap(fun):
        def deco(*args, **kwargs):
            last_trial = tries - 1
            for trial in range(tries):
                try:
                    return fun(*args, **kwargs)
                except exceptions:
                    is_last_trial = trial == last_trial

                    if retry_log is not None:
                        retry_log(trial, last_trial=is_last_trial)

                    if is_last_trial:
                        raise

                    duration = backoff_factor * 2 ** (tries - 1)

                    check_succeeds = sleep(duration, check)
                    if check_succeeds:
                        raise RetryAbortedByCheck()

        deco.__name__ = fun.__name__

        return deco

    return wrap
